@@ -5,19 +5,22 @@ Colton J. Provias
 MIT License
 """
 
-from collections import MutableMapping
-from enum import Enum
-from inflection import pluralize, dasherize, parameterize, tableize, underscore
+try:
+    from enum import Enum
+except ImportError:
+    from enum34 import Enum
+
+from inflection import dasherize, tableize, underscore
 from sqlalchemy.exc import IntegrityError
 from sqlalchemy.orm.interfaces import MANYTOONE
 from sqlalchemy.util.langhelpers import iterate_attributes
-from pprint import pprint
 
 from .errors import (BadRequestError, InvalidTypeForEndpointError,
                      MissingTypeError, NotSortableError, PermissionDeniedError,
                      RelationshipNotFoundError, ResourceNotFoundError,
                      ResourceTypeNotFoundError, ToManyExpectedError,
                      ValidationError)
+from ._version import __version__
 
 
 class AttributeActions(Enum):
@@ -44,6 +47,12 @@ class Permissions(Enum):
     EDIT = 102
     DELETE = 103
 
+class MissingKey:
+    def __init__(self, elem):
+        self.elem = elem
+
+    def __repr__(self):
+        return '<{} elem={}>'.format(self.__class__.__name__, self.elem)
 
 ALL_PERMISSIONS = {
     Permissions.VIEW, Permissions.CREATE, Permissions.EDIT, Permissions.DELETE
@@ -128,6 +137,7 @@ class PermissionTest(object):
         fn.__jsonapi_check_permission__ |= set(self.permission)
         return fn
 
+
 #: More consistent name for the decorators
 permission_test = PermissionTest
 
@@ -140,7 +150,7 @@ class JSONAPIResponse(object):
         self.status_code = 200
         self.data = {
             'jsonapi': {'version': '1.0'},
-            'meta': {'sqlalchemy_jsonapi_version': '4.0.8'}
+            'meta': {'sqlalchemy_jsonapi_version': __version__}
         }
 
 
@@ -227,7 +237,8 @@ class JSONAPI(object):
                 continue
 
             prepped_name = self._api_type_for_model(model)
-            api_type = getattr(model, '__jsonapi_type_override__', prepped_name)
+            api_type = getattr(
+                model, '__jsonapi_type_override__', prepped_name)
 
             model_keys = set(model.__mapper__.all_orm_descriptors.keys())
             model_keys |= set(model.__mapper__.relationships.keys())
@@ -236,8 +247,10 @@ class JSONAPI(object):
             model.__jsonapi_rel_desc__ = {}
             model.__jsonapi_permissions__ = {}
             model.__jsonapi_type__ = api_type
-            model.__jsonapi_map_to_py__ = {dasherize(underscore(x)): x for x in model_keys}
-            model.__jsonapi_map_to_api__ = {x: dasherize(underscore(x)) for x in model_keys}
+            model.__jsonapi_map_to_py__ = {
+                dasherize(underscore(x)): x for x in model_keys}
+            model.__jsonapi_map_to_api__ = {
+                x: dasherize(underscore(x)) for x in model_keys}
 
             for prop_name, prop_value in iterate_attributes(model):
 
@@ -259,7 +272,7 @@ class JSONAPI(object):
                     }
                     rels_desc = model.__jsonapi_rel_desc__
                     for relationship in prop_value.__jsonapi_desc_for_rels__:
-                        rels_desc.setdefault(attribute, defaults)
+                        rels_desc.setdefault(relationship, defaults)
                         rel_desc = rels_desc[relationship]
                         for action in prop_value.__jsonapi_action__:
                             rel_desc[action] = prop_value
@@ -292,8 +305,10 @@ class JSONAPI(object):
 
     def _lazy_relationship(self, api_type, obj_id, rel_key):
         return {
-            'self': '{}/{}/{}/relationships/{}'.format(self.prefix, api_type, obj_id, rel_key),
-            'related': '{}/{}/{}/{}'.format(self.prefix, api_type, obj_id, rel_key)
+            'self': '{}/{}/{}/relationships/{}'.format(
+                self.prefix, api_type, obj_id, rel_key),
+            'related': '{}/{}/{}/{}'.format(
+                self.prefix, api_type, obj_id, rel_key)
         }
 
     def _get_relationship(self, resource, rel_key, permission):
@@ -359,7 +374,9 @@ class JSONAPI(object):
         }
         attrs_to_ignore = {'__mapper__', 'id'}
         if api_type in fields.keys():
-            local_fields = list(map((lambda x: instance.__jsonapi_map_to_py__[x]), fields[api_type]))
+            local_fields = list(map((
+                lambda x: instance.__jsonapi_map_to_py__[x]),
+                fields[api_type]))
         else:
             local_fields = orm_desc_keys
 
@@ -384,23 +401,27 @@ class JSONAPI(object):
                 if api_key in include.keys():
                     related = desc(instance)
                     if related is not None:
-                        perm = get_permission_test(related, None, Permissions.VIEW)
-                    if key in local_fields and (related is None or not perm(related)):
+                        perm = get_permission_test(
+                            related, None, Permissions.VIEW)
+                    if (key in local_fields
+                            and (related is None or not perm(related))):
                         to_ret['relationships'][api_key]['data'] = None
                         continue
                     if key in local_fields:
-                        to_ret['relationships'][api_key]['data'] = self._render_short_instance(related)
+                        to_ret['relationships'][api_key]['data'] = self._render_short_instance(related)  # NOQA
                     new_include = self._parse_include(include[api_key])
-                    built = self._render_full_resource(related, new_include, fields)
+                    built = self._render_full_resource(
+                        related, new_include, fields)
                     included = built.pop('included')
                     to_ret['included'].update(included)
-                    to_ret['included'][(related.__jsonapi_type__, related.id)] = built
+                    to_ret['included'][(related.__jsonapi_type__, related.id)] = built  # NOQA
 
             else:
 
                 if key in local_fields:
                     to_ret['relationships'][api_key] = {
-                        'links': self._lazy_relationship(api_type, instance.id, api_key),
+                        'links': self._lazy_relationship(
+                            api_type, instance.id, api_key),
                     }
 
                 if api_key not in include.keys():
@@ -418,20 +439,21 @@ class JSONAPI(object):
                         continue
 
                     if key in local_fields:
-                        to_ret['relationships'][api_key]['data'].append(self._render_short_instance(item))
+                        to_ret['relationships'][api_key]['data'].append(
+                            self._render_short_instance(item))
 
                     new_include = self._parse_include(include[api_key])
                     built = self._render_full_resource(item, new_include,
                                                        fields)
                     included = built.pop('included')
                     to_ret['included'].update(included)
-                    to_ret['included'][(item.__jsonapi_type__, item.id)] = built
+                    to_ret['included'][(item.__jsonapi_type__, item.id)] = built  # NOQA
 
         for key in set(orm_desc_keys) - attrs_to_ignore:
             try:
                 desc = get_attr_desc(instance, key, AttributeActions.GET)
                 if key in local_fields:
-                    to_ret['attributes'][instance.__jsonapi_map_to_api__[key]] = desc(instance)
+                    to_ret['attributes'][instance.__jsonapi_map_to_api__[key]] = desc(instance)  # NOQA
             except PermissionDeniedError:
                 continue
 
@@ -448,7 +470,6 @@ class JSONAPI(object):
             check_permission(instance, rel_key, Permissions.EDIT)
 
             if rel.cascade.delete:
-
                 if rel.direction == MANYTOONE:
                     related = getattr(instance, rel_key)
                     self._check_instance_relationships_for_delete(related)
@@ -548,7 +569,7 @@ class JSONAPI(object):
             return ToManyExpectedError(model, resource, relationship)
 
         response = JSONAPIResponse()
-        response.data = []
+        response.data = {'data': []}
 
         session.add(resource)
 
@@ -635,6 +656,7 @@ class JSONAPI(object):
 
             attr = getattr(model, attr_name)
             if not hasattr(attr, 'asc'):
+                # pragma: no cover
                 return NotSortableError(model, attr_name)
 
             check_permission(model, attr_name, Permissions.VIEW)
@@ -756,7 +778,7 @@ class JSONAPI(object):
                                RelationshipActions.GET)(resource)
 
         if relationship.direction == MANYTOONE:
-            if related == None:
+            if related is None:
                 response.data['data'] = None
             else:
                 try:
@@ -801,7 +823,7 @@ class JSONAPI(object):
         try:
             if relationship.direction == MANYTOONE:
                 if not isinstance(json_data['data'], dict)\
-                        and json_data['data'] != None:
+                        and json_data['data'] is not None:
                     raise ValidationError('Provided data must be a hash.')
 
                 related = getattr(resource, relationship.key)
@@ -811,7 +833,7 @@ class JSONAPI(object):
                 setter = get_rel_desc(resource, relationship.key,
                                       RelationshipActions.SET)
 
-                if json_data['data'] == None:
+                if json_data['data'] is None:
                     setter(resource, None)
                 else:
                     to_relate = self._fetch_resource(
@@ -884,12 +906,13 @@ class JSONAPI(object):
         json_data['data'].setdefault('relationships', {})
         json_data['data'].setdefault('attributes', {})
 
-        data_keys = set(map((lambda x: resource.__jsonapi_map_to_py__.get(x, None)), json_data['data'].get('relationships', {}).keys()))
-        model_keys = set(resource.__mapper__.relationships.keys())
-        if not data_keys <= model_keys:
+        missing_keys = set(json_data['data'].get('relationships', {}).keys()) \
+            - set(resource.__jsonapi_map_to_py__.keys())
+
+        if missing_keys:
             raise BadRequestError(
                 '{} not relationships for {}.{}'.format(
-                    ', '.join(list(data_keys - model_keys)),
+                    ', '.join(list(missing_keys)),
                     model.__jsonapi_type__, resource.id))
 
         attrs_to_ignore = {'__mapper__', 'id'}
@@ -908,7 +931,9 @@ class JSONAPI(object):
                     session, json_data['data']['relationships'][api_key],
                     model.__jsonapi_type__, resource.id, api_key)
 
-            data_keys = set(map((lambda x: resource.__jsonapi_map_to_py__.get(x, None)), json_data['data']['attributes'].keys()))
+            data_keys = set(map((
+                lambda x: resource.__jsonapi_map_to_py__.get(x, None)),
+                json_data['data']['attributes'].keys()))
             model_keys = set(orm_desc_keys) - attrs_to_ignore
 
             if not data_keys <= model_keys:
@@ -919,12 +944,13 @@ class JSONAPI(object):
 
             for key in data_keys & model_keys:
                 setter = get_attr_desc(resource, key, AttributeActions.SET)
-                setter(resource, json_data['data']['attributes'][resource.__jsonapi_map_to_api__[key]])
+                setter(resource, json_data['data']['attributes'][resource.__jsonapi_map_to_api__[key]])  # NOQA
             session.commit()
         except IntegrityError as e:
             session.rollback()
             raise ValidationError(str(e.orig))
         except AssertionError as e:
+            # pragma: no cover
             session.rollback()
             raise ValidationError(e.msg)
         except TypeError as e:
@@ -959,13 +985,16 @@ class JSONAPI(object):
         data['data'].setdefault('relationships', {})
         data['data'].setdefault('attributes', {})
 
-        data_keys = set(map((lambda x: resource.__jsonapi_map_to_py__.get(x, None)), data['data'].get('relationships', {}).keys()))
+        data_keys = set(map((
+            lambda x: resource.__jsonapi_map_to_py__.get(x, MissingKey(x))),
+            data['data'].get('relationships', {}).keys()))
         model_keys = set(resource.__mapper__.relationships.keys())
         if not data_keys <= model_keys:
+            data_keys = set([key.elem if isinstance(key, MissingKey) else key for key in data_keys])
+            # pragma: no cover
             raise BadRequestError(
-                '{} not relationships for {}'.format(
-                    ', '.join(list(data_keys -
-                                   model_keys)), model.__jsonapi_type__))
+                    '{} not relationships for {}'.format(
+                    ', '.join([repr(key) for key in list(data_keys - model_keys)]), model.__jsonapi_type__))
 
         attrs_to_ignore = {'__mapper__', 'id'}
 
@@ -1020,9 +1049,10 @@ class JSONAPI(object):
                         raise BadRequestError(
                             '{} must be an array'.format(key))
                     for item in data_rel:
-                        if not {'type', 'id'} in set(item.keys()):
+                        if 'type' not in item.keys() or 'id' not in item.keys():
                             raise BadRequestError(
                                 '{} must have type and id keys'.format(key))
+                        # pragma: no cover
                         to_relate = self._fetch_resource(session, item['type'],
                                                          item['id'],
                                                          Permissions.EDIT)
@@ -1035,7 +1065,9 @@ class JSONAPI(object):
                                              Permissions.CREATE)
                         setters.append([setter, to_relate])
 
-            data_keys = set(map((lambda x: resource.__jsonapi_map_to_py__.get(x, None)), data['data'].get('attributes', {}).keys()))
+            data_keys = set(map((
+                lambda x: resource.__jsonapi_map_to_py__.get(x, None)),
+                data['data'].get('attributes', {}).keys()))
             model_keys = set(orm_desc_keys) - attrs_to_ignore
 
             if not data_keys <= model_keys:
@@ -1060,6 +1092,7 @@ class JSONAPI(object):
             session.rollback()
             raise ValidationError(str(e.orig))
         except AssertionError as e:
+            # pragma: no cover
             session.rollback()
             raise ValidationError(e.msg)
         except TypeError as e:
